import isFunction from "lodash/isFunction";
import { getPackageWithoutVersion, getPackageBundleUri, withRetry } from "./utils/utils";
import pickBy from "lodash/pickBy";
import { emptyConfigure } from "./components/empty-configure";
window["pieHelpers"] = {
    loadingScripts: {}
};
/**
 * TODO: Clean up: PD-761
 */
export const DEFAULT_ENDPOINTS = {
    prod: {
        bundleBase: "https://pits-cdn.pie-api.io/bundles/",
        buildServiceBase: "https://proxy.pie-api.com/bundles/",
    },
    // this is actually not really used anymore? equals to dev
    stage: {
        bundleBase: "https://pits-cdn.pie-api.io/bundles/",
        buildServiceBase: "https://proxy.pie-api.com/bundles/",
    },
    dev: {
        bundleBase: "https://pits-cdn.pie-api.io/bundles/",
        buildServiceBase: "https://proxy.dev.pie-api.com/bundles/",
    }
};
export var Status;
(function (Status) {
    Status["loading"] = "loading";
    Status["loaded"] = "loaded";
})(Status || (Status = {}));
export var BundleType;
(function (BundleType) {
    BundleType["player"] = "player.js";
    BundleType["clientPlayer"] = "client-player.js";
    BundleType["editor"] = "editor.js";
})(BundleType || (BundleType = {}));
export const needToLoad = (registry, bundle) => (el, key) => {
    if (!registry) {
        return true;
    }
    const regEntry = registry[key];
    if (!regEntry) {
        return true;
    }
    const { config, controller, element } = regEntry;
    switch (bundle) {
        case BundleType.editor:
            return !config || !controller || !element;
        case BundleType.clientPlayer:
            return !controller || !element;
        case BundleType.player:
            return !element;
        default:
            true;
    }
};
/**
 * Pie Registry contols the loading of all PIEs from the pie build service
 */
export class PieLoader {
    /**
     * Create a PieLoader instance
     * Can provide config object for endpoints, will default to to production environment preset.
     * @param endpoints an object providing {bundleBase, buildServiceBase} endpoints
     */
    constructor(_endpoints) {
        this.getController = (pieTagName) => {
            return this.registry[pieTagName]
                ? this.registry[pieTagName].controller
                : null;
        };
        this.elementsHaveLoaded = (els) => {
            const promises = els.map(el => customElements.whenDefined(el.tag));
            return Promise.all(promises)
                .then(() => {
                return Promise.resolve({ elements: els, val: true });
            })
                .catch(() => {
                return Promise.resolve({ elements: els, val: false });
            });
        };
        this.getBaseUrls = (options, piesToLoad) => {
            const bundleUri = getPackageBundleUri(piesToLoad);
            if (!bundleUri) {
                return;
            }
            return (options.endpoints.buildServiceBase + bundleUri + "/" + options.bundle);
        };
        this.getScriptsUrl = (options, piesToLoad) => {
            if (options.forceBundleUrl) {
                const url = this.getBaseUrls(options, piesToLoad);
                if (url) {
                    return url;
                }
            }
            if (options.content.bundle && options.content.bundle.url) {
                return options.content.bundle.url;
            }
            if (options.useCdn &&
                options.content.bundle &&
                options.content.bundle.hash) {
                return `${options.endpoints.bundleBase + options.content.bundle.hash}/${options.bundle}`;
            }
            return this.getBaseUrls(options, piesToLoad);
        };
        /**
         *
         * @param {Object<string,string>} elements elements to load from pie cloud service
         * @param {HTMLDocument} doc - the document to load the scripts
         * @param {string} base_url - default base url for cloud service
         */
        this.loadCloudPies = async (options) => {
            performance.mark('pie-loader-start');
            if (!options.endpoints) {
                options.endpoints = this.endpoints;
            }
            if (!options.bundle) {
                options.bundle = BundleType.editor;
            }
            const elements = options.content.elements;
            let head = options.doc.getElementsByTagName("head")[0];
            if (!head) {
                head = options.doc.createElement("head");
                options.doc.appendChild(head);
            }
            performance.mark('pie-getElementsToLoad-start');
            const piesToLoad = this.getElementsToLoad(elements, options.bundle, this.registry);
            performance.mark('pie-getElementsToLoad-end');
            performance.measure('PIE Get Elements To Load Time', 'pie-getElementsToLoad-start', 'pie-getElementsToLoad-end');
            let scriptUrl = this.getScriptsUrl(options, piesToLoad);
            if (!scriptUrl) {
                console.error("No script urls found for elements.");
                return;
            }
            // These scripts are actually being added to headers ONLY when the fetch below (in loadScript) returns a value
            // However, if we have multiple players that are supposed to load the same bundle on the same page,
            //  this means that we'll make that fetch request multiple times, which slows down the page
            const loadedScripts = [...head.getElementsByTagName("script")];
            // That's why we're using this little helper to store the ones that are in the process of loading as well
            const alreadyLoadingScript = window["pieHelpers"] && window["pieHelpers"].loadingScripts[scriptUrl];
            if (loadedScripts.find(s => (s.src === scriptUrl)) || alreadyLoadingScript) {
                return;
            }
            if (window["pieHelpers"] && !window["pieHelpers"].loadingScripts[scriptUrl]) {
                window["pieHelpers"].loadingScripts[scriptUrl] = true;
            }
            const script = options.doc.createElement("script");
            const onloadFn = (_pies => {
                return () => {
                    performance.mark('pie-loader-end');
                    performance.measure('PIE Loader Bundle Time', 'pie-loader-start', 'pie-loader-end');
                    const entry = performance.getEntriesByName('PIE Loader Bundle Time')[0];
                    const duration = entry ? entry.duration : 0;
                    console.log('[PIE Loader Bundle Time]', duration.toFixed(2), 'ms');
                    performance.mark('pie-custom-elements-define-start');
                    const pieKeys = Object.keys(_pies);
                    pieKeys.forEach(key => {
                        const packagesWithoutVersion = getPackageWithoutVersion(_pies[key]);
                        const pie = window["pie"] && window["pie"].default
                            ? window["pie"].default[packagesWithoutVersion]
                            : null;
                        if (!pie) {
                            console.log(`pie constructor not found for ${key}`);
                            return;
                        }
                        const elName = key;
                        if (!customElements.get(elName)) {
                            customElements.define(elName, pie.Element);
                            this.registry[elName] = {
                                package: _pies[key],
                                status: Status.loading,
                                tagName: elName
                            };
                            customElements.whenDefined(elName).then(async () => {
                                this.registry[elName].status = Status.loaded;
                                this.registry[elName].element = customElements.get(elName);
                                this.registry[elName].controller = pie.controller;
                            });
                        }
                        if (options.bundle === BundleType.editor) {
                            // This fixes some cases where the pie build service fails
                            pie.Configure = isFunction(pie.Configure)
                                ? pie.Configure
                                : emptyConfigure(elName);
                            const configElName = elName + "-config";
                            if (!customElements.get(configElName)) {
                                customElements.define(configElName, pie.Configure);
                                customElements.whenDefined(configElName).then(async () => {
                                    if (this.registry[elName]) {
                                        this.registry[elName].config = customElements.get(configElName);
                                    }
                                });
                            }
                        }
                    });
                    performance.mark('pie-custom-elements-define-end');
                    performance.measure('PIE Define Custom Elements Time', 'pie-custom-elements-define-start', 'pie-custom-elements-define-end');
                    const defineEntry = performance.getEntriesByName('PIE Define Custom Elements Time')[0];
                    const defineDuration = defineEntry ? defineEntry.duration : 0;
                    console.log('[PIE Define Custom Elements Time]', defineDuration.toFixed(2), 'ms');
                };
            })(piesToLoad);
            if (options.reFetchBundle) {
                const loadScript = async () => {
                    try {
                        const response = await withRetry(async (currentDelay) => {
                            const res = await fetch(scriptUrl);
                            // if the request fails with 503 retry it
                            if (res.status === 503) {
                                console.warn(`Service unavailable (503), retrying in ${currentDelay / 1000 ||
                                    1} seconds...`);
                                throw new Error("Unavailable, retrying");
                            }
                            return res;
                        }, 20, 1000, 30000);
                        // if the request is successful inject the response as a script tag
                        // to avoid doing the same call twice
                        if (response.status === 200) {
                            script.onload = onloadFn;
                            script.src = scriptUrl;
                            head.appendChild(script);
                            delete window["pieHelpers"].loadingScripts[scriptUrl];
                        }
                        else {
                            console.error("Failed to load script, status code:", response.status);
                        }
                    }
                    catch (error) {
                        console.error("Network error occurred while trying to load script:", error);
                    }
                };
                await loadScript();
            }
            else {
                script.onload = onloadFn;
                script.src = scriptUrl;
                head.appendChild(script);
            }
        };
        /**
         * Given a defintion of elements, will check the registry
         * and return the elements and tags that need to be loaded.
         *
         * @param elements - the elements to test against registry
         */
        this.getElementsToLoad = (els, bundle = BundleType.editor, registry) => {
            return pickBy(els, needToLoad(registry, bundle));
        };
        if (!_endpoints) {
            this.endpoints = DEFAULT_ENDPOINTS.prod;
        }
        else {
            this.endpoints = _endpoints;
        }
        // read from global in case >1 instance loaded.
        window["PIE_REGISTRY"] = window["PIE_REGISTRY"]
            ? window["PIE_REGISTRY"]
            : {};
        PieLoader._registry = window["PIE_REGISTRY"];
        this.registry = PieLoader._registry;
    }
}
